'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.buildSchemaFromTypeDefinitions = exports.addTracingToResolvers = exports.assertResolveFunctionsPresent = exports.addCatchUndefinedToSchema = exports.addResolveFunctionsToSchema = exports.addErrorLoggingToSchema = exports.forEachField = exports.SchemaError = exports.generateSchema = undefined;

var _language = require('graphql/language');

var _utilities = require('graphql/utilities');

var _type = require('graphql/type');

var _tracing = require('./tracing');

// @schemaDefinition: A GraphQL type schema in shorthand
// @resolvers: Definitions for resolvers to be merged with schema
// Generates a schema for graphql-js given a shorthand schema

function SchemaError(message) {
  Error.captureStackTrace(this, this.constructor);
  this.message = message;
}
SchemaError.prototype = new Error();

var generateSchema = function generateSchema(typeDefinitions, resolveFunctions) {
  var logger = arguments.length <= 2 || arguments[2] === undefined ? null : arguments[2];
  var forbidUndefinedInResolve = arguments.length <= 3 || arguments[3] === undefined ? false : arguments[3];

  if (!typeDefinitions) {
    throw new SchemaError('Must provide typeDefinitions');
  }
  if (!resolveFunctions) {
    throw new SchemaError('Must provide resolveFunctions');
  }

  var schema = buildSchemaFromTypeDefinitions(typeDefinitions);

  addResolveFunctionsToSchema(schema, resolveFunctions);

  assertResolveFunctionsPresent(schema);

  if (forbidUndefinedInResolve) {
    addCatchUndefinedToSchema(schema);
  }

  if (logger) {
    addErrorLoggingToSchema(schema, logger);
  }

  return schema;
};

function buildSchemaFromTypeDefinitions(typeDefinitions) {
  return (0, _utilities.buildASTSchema)((0, _language.parse)(typeDefinitions));
}

function forEachField(schema, fn) {
  var typeMap = schema.getTypeMap();
  Object.keys(typeMap).forEach(function (typeName) {
    var type = typeMap[typeName];

    if (!(0, _type.getNamedType)(type).name.startsWith('__') && type instanceof _type.GraphQLObjectType) {
      (function () {
        var fields = type.getFields();
        Object.keys(fields).forEach(function (fieldName) {
          var field = fields[fieldName];
          fn(field, typeName, fieldName);
        });
      })();
    }
  });
}

function addResolveFunctionsToSchema(schema, resolveFunctions) {
  Object.keys(resolveFunctions).forEach(function (typeName) {
    var type = schema.getType(typeName);
    if (!type) {
      throw new SchemaError('"' + typeName + '" defined in resolvers, but not in schema');
    }

    Object.keys(resolveFunctions[typeName]).forEach(function (fieldName) {
      if (!type.getFields()[fieldName]) {
        throw new SchemaError(typeName + '.' + fieldName + ' defined in resolvers, but not in schema');
      }
      var field = type.getFields()[fieldName];
      field.resolve = resolveFunctions[typeName][fieldName];
    });
  });
}

function assertResolveFunctionsPresent(schema) {
  forEachField(schema, function (field, typeName, fieldName) {
    // requires a resolve function on every field that has arguments
    if (field.args.length > 0) {
      expectResolveFunction(field, typeName, fieldName);
    }

    // requires a resolve function on every field that returns a non-scalar type
    if (!((0, _type.getNamedType)(field.type) instanceof _type.GraphQLScalarType)) {
      expectResolveFunction(field, typeName, fieldName);
    }
  });
}

function expectResolveFunction(field, typeName, fieldName) {
  if (!field.resolve) {
    throw new SchemaError('Resolve function missing for "' + typeName + '.' + fieldName + '"');
  }
  if (typeof field.resolve !== 'function') {
    throw new SchemaError('Resolver "' + typeName + '.' + fieldName + '" must be a function');
  }
}

function addErrorLoggingToSchema(schema, logger) {
  if (!logger) {
    throw new Error('Must provide a logger');
  }
  if (typeof logger.log !== 'function') {
    throw new Error('Logger.log must be a function');
  }
  forEachField(schema, function (field, typeName, fieldName) {
    var errorHint = typeName + '.' + fieldName;
    // eslint-disable-next-line no-param-reassign
    field.resolve = decorateWithLogger(field.resolve, logger, errorHint);
  });
}

/*
 * fn: The function to decorate with the logger
 * logger: an object instance of type Logger
 * hint: an optional hint to add to the error's message
 */
function decorateWithLogger(fn, logger) {
  var hint = arguments.length <= 2 || arguments[2] === undefined ? '' : arguments[2];

  return function () {
    try {
      return fn.apply(undefined, arguments);
    } catch (e) {
      if (hint) {
        e.message = 'Error in resolver ' + hint + '\n' + e.message;
      }
      logger.log(e);
      // we want to pass on the error, just in case.
      throw e;
    }
  };
}

function addCatchUndefinedToSchema(schema) {
  forEachField(schema, function (field, typeName, fieldName) {
    var errorHint = typeName + '.' + fieldName;
    // eslint-disable-next-line no-param-reassign
    field.resolve = decorateToCatchUndefined(field.resolve, errorHint);
  });
}

function addTracingToResolvers(schema, tracer) {
  forEachField(schema, function (field, typeName, fieldName) {
    var functionName = typeName + '.' + fieldName;
    // eslint-disable-next-line no-param-reassign
    field.resolve = (0, _tracing.decorateWithTracer)(field.resolve, tracer, { functionType: 'resolve', functionName: functionName });
  });
}

function decorateToCatchUndefined(fn, hint) {
  return function () {
    var result = fn.apply(undefined, arguments);
    if (typeof result === 'undefined') {
      throw new Error('Resolve function for "' + hint + '" returned undefined');
    }
    return result;
  };
}

exports.generateSchema = generateSchema;
exports.SchemaError = SchemaError;
exports.forEachField = forEachField;
exports.addErrorLoggingToSchema = addErrorLoggingToSchema;
exports.addResolveFunctionsToSchema = addResolveFunctionsToSchema;
exports.addCatchUndefinedToSchema = addCatchUndefinedToSchema;
exports.assertResolveFunctionsPresent = assertResolveFunctionsPresent;
exports.addTracingToResolvers = addTracingToResolvers;
exports.buildSchemaFromTypeDefinitions = buildSchemaFromTypeDefinitions;